

/*
 *
 *	This file is similar to /usr/include/regexp.h except that
 *	it has been modified to use the Dyalog apl character set.
 *	There are no signed characters used.
 *	This file depends on quadav.h
 *
 *	An extra parameter has been added to each of the functions 
 *	step() and advance()
 *
 *	The incoming string is not zero terminated. Instead the 
 *	new second parameter 'elp' points to the last character
 *	in the string. (NB not one character past the last)
 */

#define	CBRA		2
#define	CCHR		4
#define	CDOT		8
#define	CCL		12
#define	CDOL		20
#define	CCEOF		22
#define	CKET		24
#define	CBACK		36

#define	ASTERISK	01
#define	RNGE		03
#define LTHAN           40
#define	GTHAN		44
#define	CLCCHR		48	// caseless char , added 13 oct 2000 cvc

#define	NBRA		9
#define EPSIZE          32

#define	PLACE(c)	ep[c >> 3] |= bittab[c & 0x7]
#define	ISTHERE(c)	(ep[c >> 3] & bittab[c & 0x7])

#define	NGETC()		(++count, GETC())
#define	NUNGETC(c)	(--count, UNGETC(c))

UCHAR	*braslist[NBRA];
UCHAR	*braelist[NBRA];
UCHAR	*loc1=NULL, *loc2=NULL, *lpstart=NULL;
int	sed;
int	nodelim;

int	circf;
int	low;
int	size;

UCHAR	bittab[] = 
		{
		1,
		2,
		4,
		8,
		16,
		32,
		64,
		128
		};

static UCHAR avtoupper(UCHAR c)
{
	if (c >= AL && c <= ZL)
		c += A-AL;
	else if (c >= DELTAU && c <= ZU)
		c -= AU-A;
	return c;
}

static UCHAR avtolower(UCHAR c)
{
	if (c >= A && c <= Z)
		c -= A-AL;
	else if (c >= AU && c <= ZU)
		c -= AU-AL;
	else if (c == DELTA)
		c = DELTAU;
	return c;
}
static UCHAR uscore(UCHAR c)
{		//xutils ap prior to october 2000 matched a, A, and AU in
		//caseless mode.  This function helps us keep that functionality
	if (c >= AL && c <= ZL)
		c += AU - AL;
	else if (c >= A && c <= Z)
		c += AU - A;
	return c;
}

int step(UCHAR *p1, UCHAR *ep1, UCHAR *p2)
{
	REG c;
	int retval;

	if(circf) 
		{
		loc1 = p1;
		return(advance(p1, ep1, p2));
		}
	// fast check for first character 
	if(*p2==CCHR) 
		{
		c = p2[1];
		do 
			{
			if(p1 <= ep1 && *p1 != c)
				continue;
			if(advance(p1, ep1, p2)) 
				{
				loc1 = p1;
				return(1);
				}
			} while(p1++ <= ep1);
		return(0);
		}
	else if(*p2==CLCCHR) 
		{
		c = p2[1];
		do 
			{
			if(p1 <= ep1 && avtoupper(*p1) != c)
				continue;
			if(advance(p1, ep1, p2)) 
				{
				loc1 = p1;
				return(1);
				}
			} while(p1++ <= ep1);
		return(0);
		}

		// regular algorithm 
	do 
		{
		if(retval=advance(p1, ep1, p2)) 
			{
			loc1 = p1+(retval==2);
			return(1);
			}
		} while(p1++ <= ep1);
	return(0);
}

getrnge(str)
REG UCHAR *str;
{
	low = *str++;
	size = *str == 255 ? 20000 :*str - low;
}

ecmp(a, b, count)
REG UCHAR	*a, *b;
REG	count;
{
	while(count--)
		if(*a++ != *b++)
			return(0);
	return(1);
}
static in_word(c)
REG UCHAR c;
{
if (c>=ZERO && c<=NINE) return 1;
if (c>=A && c <=Z) return 1;
if (c>=AL && c<=ZL) return 1;
if (c>=AU && c<=ZU) return 1;
return 0;
}

int advance(UCHAR *lp, UCHAR *elp, UCHAR *ep)
{
	REG UCHAR *curlp;
	UCHAR c;
	UCHAR *bbeg;
	int ct;

	while(1)
		switch(*ep++) 
			{

		case CCHR:
			if(lp > elp)
				return 0;

			if(*ep++ == *lp++)
				continue;
			return(0);
		case CLCCHR:
			if(lp > elp)
				return 0;

			if(*ep++ == avtoupper(*lp++))
				continue;
			return(0);
		case CDOT:
			if(lp++ <= elp)
				continue;
			return(0);

		case CDOL:
			if(lp == elp+1)
				continue;
			return(0);

		case CCEOF:
			loc2 = lp;
			return(1);
		case LTHAN:
#if 0
/* This was the code as of July 15, 1997 */			
			if (lp==lpstart) continue;
			if (in_word(*lp)) return 0;
		   // 	*--ep =0; 
			return (advance(++lp,elp,ep) ? 2 : 0) ;
			
			/*
			 * This code had problems with:
			 * finding \<word\> in 'word'
			 */
#endif
			if(lp > elp)	//added here for prob sept 2000
				return 0; 

				/*new code to fix bug 8084 */
				/* added july 22,1997, CVC */
			ep++; 	//corresponds to ep++ in compile().
			if (lp==lpstart) 
				continue;
			if (in_word(*(lp-1)) )
				return 0;
			else
				continue;
		case GTHAN:
			if (lp >=elp) continue;
			if (in_word(*lp)) return 0;
		//	*--ep=0; 
			return advance(lp,elp,ep);
		case CCL:
			if(lp > elp)
				return 0; 
			c = *lp++;
			if (ISTHERE(c) ){
					ep += EPSIZE;
					continue;
					}
			return(0);
			
		case CBRA:
			if(lp > elp)
				return 0; 
			braslist[*ep++] = lp;
			continue;

		case CKET:
			if(lp > elp)
				return 0; 
			braelist[*ep++] = lp;
			continue;

		case CCHR|RNGE:
			if(lp > elp)
				return 0; 

			c = *ep++;
			getrnge(ep);
			for(;low--;lp++)
				if(*lp != c)
					return(0);
			curlp = lp;
			for(;size--;lp++) 
				if(*lp != c)
					break;
			ep += 2;
			goto asterisk;

		case CLCCHR|RNGE:
			if(lp > elp)
				return 0; 

			c = *ep++;
			getrnge(ep);
			for(;low--;lp++)
				if(avtoupper(*lp) != c)
					return(0);

			curlp = lp;
			for(;size--;lp++) 
				if(avtoupper(*lp) != c)
					break;

			ep += 2;
			goto asterisk;


		case CDOT|RNGE:
			if(lp > elp)
				return 0; 

			getrnge(ep);
			for(;low--;lp++)
				if(lp > elp)
					return(0);
			curlp = lp;
			for(;size--;lp++)
				if(lp > elp)
					break;
			ep += 2;
			goto asterisk;

		case CCL|RNGE:
			if(lp > elp)
				return 0; 

			getrnge(ep + EPSIZE);
			for(;low--;lp++) 
				{
				c = *lp;
				if(!ISTHERE(c))
					return(0);
				}
			curlp = lp;
			for(;size--;lp++) 
				{
				c = *lp;
				if(!ISTHERE(c))
					break;
				}
			ep += EPSIZE+2;
			goto asterisk;
		case CBACK:
			bbeg = braslist[*ep];
			ct = braelist[*ep++] - bbeg;

			if(lp > elp)
				return 0; 

			if(ecmp(bbeg, lp, ct)) 
				{
				lp += ct;
				continue;
				}
			return(0);

		case CBACK|ASTERISK:
			bbeg = braslist[*ep];
			ct = braelist[*ep++] - bbeg;
			curlp = lp;

			while((lp + ct <= elp) && ecmp(bbeg, lp, ct))
					lp += ct;

			while(lp >= curlp) 
				{
				if(advance(lp, elp, ep)) return(1);
				lp -= ct;
				}
			return(0);


		case CDOT|ASTERISK:
			curlp = lp;
			while(lp <= elp)
				lp++;
			goto asterisk;

		case CCHR|ASTERISK:
			curlp = lp;
			while(lp <= elp && *lp == *ep)
				lp++;
			ep++;
			goto asterisk;
	
		case CLCCHR|ASTERISK:
			curlp = lp;
			while(lp <= elp && avtoupper(*lp) == *ep)
				lp++;
			ep++;
			goto asterisk;

		case CCL|ASTERISK:
			curlp = lp;
			for(;lp <= elp; lp++)
				{
				c = *lp;
				if (!ISTHERE(c))
					break;
				}
			ep += EPSIZE;
			goto asterisk;

		asterisk:
			do 
				{
				if(advance(lp, elp, ep))
					return(1);
				--lp;
				} while(lp >= curlp);
			return(0);

		}
}
void compile(UCHAR *instring,REG UCHAR *ep,UCHAR *endbuf, REG UNS elen, 
		int caseless) 
{
	INIT	// Dependent declarations and initializations 
	REG unsigned c;
	REG unsigned count;
	UCHAR *lastep = instring;
	int cclcnt;
	UCHAR bracket[NBRA], *bracketp;
	int closed = 0;
	UCHAR neg, *tmpptr;
	int lc;
	int i, cflg;
	int nbra=0, ebra=0;	//Fix Problem 1933 
	count = 0;
	lastep = 0;

	if(elen == 0)		// null string: can't risk NGETC().
		{
		if(*ep == 0 && !sed)
			ERROR(41);
		goto ret;
		}

	c = NGETC();
	if(c == NEWLINE) 
		{
		NUNGETC(c);
		nodelim = 1;
		if(*ep == 0 && !sed)
			ERROR(41);
		goto ret;
		}
	bracketp = bracket;
	circf = 0;
	if(c == AND)
		circf++;
	else
		NUNGETC(c);
	while(1)
		{
		if(ep >= endbuf)
			ERROR(50);

		if(count >= elen)		// none left
			{
			*ep++ = CCEOF;
			goto ret;
			}

		c = NGETC();
		if(c != STAR && ((c != SLOPE) || (PEEKC() != LBRACE)))
			lastep = ep;
		switch(c) 
			{
		case POINT:
			*ep++ = CDOT;
			continue;

		case NEWLINE:
			if(!sed) 
				{
				NUNGETC(c);
				*ep++ = CCEOF;
				nodelim = 1;
				goto ret;
				}
			else ERROR(36);
		case STAR:
			if(lastep==0 || *lastep==CBRA || *lastep==CKET)
				goto defchar;
			*lastep |= ASTERISK;
			continue;

		case DOLLAR:
			if(count < elen && PEEKC() != NEWLINE)
				goto defchar;
			*ep++ = CDOL;
			continue;
		
		case RBRACKET:
			ERROR(49);
		case LBRACKET:
			if(&ep[EPSIZE+1] >= endbuf)
				ERROR(50);

			*ep++ = CCL;
			lc = 0;
			for(i = 0; i < EPSIZE; i++)
				ep[i] = 0;

			neg = 0;
			if((c = NGETC()) == AND) 
				{
				neg = 1;
				c = NGETC();  
				}
			if (c==RBRACKET) // first el 
			{
				tmpptr = sp;
				for (i=count;i<=elen;i++)
				if (RBRACKET == *tmpptr++)
				{	
					PLACE(c);
					c=NGETC();
					break;
				}
			}
			while(RBRACKET != c)
				{
				if(c == 0 || c == NEWLINE)
					ERROR(49);
				if(c == MINUS && lc != 0) 
					{
					if(RBRACKET==(c = NGETC())) 
						{
						PLACE(MINUS);
						break;
						}
					while(lc < c) 
						{
						PLACE(lc);
						lc++;
						}
					}
				lc = c;
				if (caseless)
					{
					PLACE(avtolower(c));
					PLACE(avtoupper(c));
					PLACE(uscore(c));
					}
				else
					{
					PLACE(c);   
					}
				c=NGETC();
				}
			if(neg) 
				{
				for(cclcnt = 0; cclcnt < EPSIZE; cclcnt++)
					ep[cclcnt] ^= 0xff;
				ep[0] &= 0376;
				}

			ep += EPSIZE;

			continue;

		case SLOPE:
			switch(c = NGETC()) 
				{
			case LT:
				*ep++ = LTHAN;
				ep++;
				continue;
			case GT:
				*ep++ = GTHAN;
				continue;
			case LPAREN:
				if(nbra >= NBRA)
					ERROR(43);
				*bracketp++ = nbra;
				*ep++ = CBRA;
				*ep++ = nbra++;
				continue;
				
			case RPAREN:
				if(bracketp <= bracket || ++ebra != nbra)
					ERROR(42);
				*ep++ = CKET;
				*ep++ = *--bracketp;
				closed++;
				continue;

			case LBRACE:
				if(lastep == (UCHAR *) (0))
					goto defchar;
				*lastep |= RNGE;
				cflg = 0;
			nlim:
				c = NGETC();
				i = 0;
				do 
					{
					if(ZERO <= c && c <= NINE)
						i = 10 * i + c - ZERO;
					else
						ERROR(16);
					} while(((c = NGETC()) != SLOPE)
						&& (c != COMMA));
				if(i > 255)
					ERROR(11);
				*ep++ = i;
				if(c == COMMA) 
					{
					if(cflg++)
						ERROR(44);
					if((c = NGETC()) == SLOPE)
						*ep++ = 255;
					else 
						{
						NUNGETC(c);
						goto nlim;
						// get 2'nd number 
						}
					}
				if(NGETC() != RBRACE)
					ERROR(45);
				if(!cflg)	// one number 
					*ep++ = i;
				else if((ep[-1]) < (ep[-2]))
					ERROR(46);
				continue;

			case NEWLINE:
				ERROR(36);

			case N:
				c = NEWLINE;
				goto defchar;

			default:
				if(c >= ONE && c <= NINE) 
					{
					if((c -= ONE) >= closed)
						ERROR(25);
					*ep++ = CBACK;
					*ep++ = c;
			continue;
					}
				}
	// Drop through to default to use \ to turn off special chars 
		defchar:
		default:
			lastep = ep;
			if (caseless)
				{
				*ep++ = CLCCHR;
				*ep++ = avtoupper(c);
				}
			else
				{
				*ep++ = CCHR;
				*ep++ = c;
				}

			}
			continue;
		ret:
		if (nbra != closed) ERROR(42);
		return;
		}
}

